import pytest

from great_expectations.data_context.store import (
    ValidationResultStore,
)
from great_expectations.data_context.types import (
    ValidationResultIdentifier,
)
from great_expectations.exceptions import MissingTopLevelConfigKeyError
from great_expectations.util import (
    gen_directory_tree_str,
)

def test_ValidationResultStore_with_InMemoryStoreBackend():

    my_store = ValidationResultStore(
        store_backend = {
            "module_name" : "great_expectations.data_context.store",
            "class_name" : "InMemoryStoreBackend",
            "separator" : ".",
        },
        root_directory=None,#"dummy/path/",
    )

    with pytest.raises(TypeError):
        my_store.get("not_a_ValidationResultIdentifier")

    with pytest.raises(MissingTopLevelConfigKeyError):
        my_store.get(ValidationResultIdentifier(**{}))
    
    ns_1 = ValidationResultIdentifier(
        from_string="ValidationResultIdentifier.a.b.c.quarantine.prod-100"
    )
    my_store.set(ns_1,{"A": "aaa"})
    assert my_store.get(ns_1) == {"A": "aaa"}

    ns_2 = ValidationResultIdentifier(
        from_string="ValidationResultIdentifier.a.b.c.quarantine.prod-200"
    )
    my_store.set(ns_2, "bbb")
    assert my_store.get(ns_2) == "bbb"

    # Verify that internals are working as expected
    assert my_store.store_backend.store == {
        'a.b.c.quarantine.prod-100': '{"A": "aaa"}',
        'a.b.c.quarantine.prod-200': '"bbb"',
    }

    print(my_store.list_keys())
    assert set(my_store.list_keys()) == set([
        ns_1,
        ns_2,
    ])

def test_ValidationResultStore__convert_resource_identifier_to_list():

    my_store = ValidationResultStore(
        store_backend = {
            "module_name" : "great_expectations.data_context.store",
            "class_name" : "InMemoryStoreBackend",
        },
        root_directory=None,
    )

    ns_1 = ValidationResultIdentifier(
        from_string="ValidationResultIdentifier.a.b.c.quarantine.prod-100"
    )
    assert my_store._convert_resource_identifier_to_tuple(ns_1) == ('a', 'b', 'c', 'quarantine', 'prod-100')
    
def test_ValidationResultStore_with_FixedLengthTupleFileSystemStoreBackend(tmp_path_factory):
    path = str(tmp_path_factory.mktemp('test_ValidationResultStore_with_FixedLengthTupleFileSystemStoreBackend__dir'))
    project_path = str(tmp_path_factory.mktemp('my_dir'))

    my_store = ValidationResultStore(
        store_backend = {
            "module_name" : "great_expectations.data_context.store",
            "class_name" : "FixedLengthTupleFilesystemStoreBackend",
            "base_directory" : "my_store/",
            "filepath_template" : "{4}/{0}/{1}/{2}/{3}.txt",
        },
        root_directory = path,
    )

    with pytest.raises(TypeError):
        my_store.get("not_a_ValidationResultIdentifier")

    with pytest.raises(MissingTopLevelConfigKeyError):
        my_store.get(ValidationResultIdentifier(**{}))
    
    ns_1 = ValidationResultIdentifier(
        from_string="ValidationResultIdentifier.a.b.c.quarantine.prod-100"
    )
    my_store.set(ns_1, {"A": "aaa"})
    assert my_store.get(ns_1) == {"A": "aaa"}

    ns_2 = ValidationResultIdentifier(
        from_string="ValidationResultIdentifier.a.b.c.quarantine.prod-20"
    )
    my_store.set(ns_2, {"B": "bbb"})
    assert my_store.get(ns_2) == {"B": "bbb"}

    print(my_store.list_keys())
    assert set(my_store.list_keys()) == set([
        ns_1,
        ns_2,
    ])

    print(gen_directory_tree_str(path))
    assert gen_directory_tree_str(path) == """\
test_ValidationResultStore_with_FixedLengthTupleFileSystemStoreBackend__dir0/
    my_store/
        prod-100/
            a/
                b/
                    c/
                        quarantine.txt
        prod-20/
            a/
                b/
                    c/
                        quarantine.txt
"""



"""
Tests for autoinspection framework.
"""

import pytest
from .test_utils import get_dataset
from .conftest import CONTEXTS

import great_expectations as ge


def test_no_autoinspection():
    df = ge.dataset.PandasDataset({"a": [1, 2, 3]}, profiler=None)
    suite = df.get_expectation_suite()

    assert len(suite["expectations"]) == 0


def test_default_no_autoinspection():
    df = ge.dataset.PandasDataset({"a": [1, 2, 3]})
    suite = df.get_expectation_suite()

    assert len(suite["expectations"]) == 0


@pytest.mark.parametrize("dataset_type", CONTEXTS)
def test_autoinspect_existing_dataset(dataset_type):
    # Get a basic dataset with no expectations
    df = get_dataset(dataset_type, {"a": [1, 2, 3]}, profiler=None)
    suite = df.get_expectation_suite()
    assert len(suite["expectations"]) == 0

    # Run autoinspect
    df.profile(ge.profile.ColumnsExistProfiler)
    suite = df.get_expectation_suite()

    # Ensure that autoinspect worked
    assert suite["expectations"] == \
        [{'expectation_type': 'expect_column_to_exist', 'kwargs': {'column': 'a'}}]


@pytest.mark.parametrize("dataset_type", CONTEXTS)
def test_autoinspect_columns_exist(dataset_type):
    df = get_dataset(
        dataset_type, {"a": [1, 2, 3]}, profiler=ge.profile.ColumnsExistProfiler)
    suite = df.get_expectation_suite()

    assert len(suite["expectations"]) == 1
    assert suite["expectations"] == \
        [{'expectation_type': 'expect_column_to_exist', 'kwargs': {'column': 'a'}}]


def test_autoinspect_warning():
    with pytest.raises(NotImplementedError):
        ge.dataset.Dataset(profiler=ge.profile.ColumnsExistProfiler)



from .datasource import Datasource
from .pandas_source import PandasDatasource
from .sqlalchemy_source import SqlAlchemyDatasource
from .spark_source import SparkDFDatasource
from .dbt_source import DBTDatasource



# -*- coding: utf-8 -*-

import os
import glob
import shutil

from great_expectations.data_context.util import safe_mmkdir
from great_expectations import __version__ as ge_version


def file_relative_path(dunderfile, relative_path):
    """
    This function is useful when one needs to load a file that is
    relative to the position of the current file. (Such as when
    you encode a configuration file path in source file and want
    in runnable in any current working directory)

    It is meant to be used like the following:
    file_relative_path(__file__, 'path/relative/to/file')

    H/T https://github.com/dagster-io/dagster/blob/8a250e9619a49e8bff8e9aa7435df89c2d2ea039/python_modules/dagster/dagster/utils/__init__.py#L34
    """
    return os.path.join(os.path.dirname(dunderfile), relative_path)


def script_relative_path(file_path):
    """
    Useful for testing with local files. Use a path relative to where the
    test resides and this function will return the absolute path
    of that file. Otherwise it will be relative to script that
    ran the test

    Note: this is function is very, very expensive (on the order of 1
    millisecond per invocation) so this should only be used in performance
    insensitive contexts. Prefer file_relative_path for anything with
    performance constraints.
    """
    # from http://bit.ly/2snyC6s

    import inspect
    scriptdir = inspect.stack()[1][1]
    return os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(scriptdir)), file_path))


def scaffold_directories_and_notebooks(base_dir):
    """Add basic directories for an initial, opinionated GE project."""

    safe_mmkdir(base_dir, exist_ok=True)
    notebook_dir_name = "notebooks"

    open(os.path.join(base_dir, ".gitignore"), 'w').write("uncommitted/")

    for directory in [notebook_dir_name, "expectations", "datasources", "uncommitted", "plugins", "fixtures"]:
        safe_mmkdir(os.path.join(base_dir, directory), exist_ok=True)

    for uncommitted_directory in ["validations", "documentation", "samples"]:
        safe_mmkdir(os.path.join(base_dir, "uncommitted",
                                 uncommitted_directory), exist_ok=True)

    for notebook in glob.glob(file_relative_path(__file__, "../init_notebooks/*.ipynb")):
        notebook_name = os.path.basename(notebook)
        shutil.copyfile(notebook, os.path.join(
            base_dir, notebook_dir_name, notebook_name))


# !!! This injects a version tag into the docs. We should test that those versioned docs exist in RTD.
greeting_1 = """
Always know what to expect from your data.

If you're new to Great Expectations, this tutorial is a good place to start:

    <blue>https://docs.greatexpectations.io/en/latest/getting_started/cli_init.html?utm_source=cli&utm_medium=init&utm_campaign={0:s}</blue>
""".format(ge_version.replace(".", "_"))

msg_prompt_lets_begin = """
Let's add Great Expectations to your project, by scaffolding a new great_expectations directory:

    great_expectations
    ├── .gitignore
    ├── datasources
    ├── expectations
    ├── fixtures
    ├── great_expectations.yml
    ├── notebooks
    │   ├── create_expectations.ipynb
    │   └── integrate_validation_into_pipeline.ipynb
    ├── plugins
    └── uncommitted
        ├── config_variables.yml
        ├── documentation
        │   ├── local_site
        │   └── team_site
        ├── samples
        └── validations
    
OK to proceed?
"""



import requests
import datetime
import logging
import os
import json
import errno
import six
import importlib
import copy
import re
from collections import OrderedDict

logger = logging.getLogger(__name__)


def safe_mmkdir(directory, exist_ok=True):
    """Simple wrapper since exist_ok is not available in python 2"""
    if not isinstance(directory, six.string_types):
        raise TypeError("directory must be of type str, not {0}".format({
            "directory_type": str(type(directory))
        }))

    if not exist_ok:
        raise ValueError(
            "This wrapper should only be used for exist_ok=True; it is designed to make porting easier later")
    try:
        os.makedirs(directory)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


# TODO : Consider moving this into types.resource_identifiers.DataContextKey.
# NOTE : We **don't** want to encourage stringification of keys, other than in tests, etc.
# TODO : Rename to parse_string_to_data_context_key
def parse_string_to_data_context_resource_identifier(string, separator="."):
    string_elements = string.split(separator)

    loaded_module = importlib.import_module("great_expectations.data_context.types.resource_identifiers")
    class_ = getattr(loaded_module, string_elements[0])

    class_instance = class_(*(string_elements[1:]))

    return class_instance


def load_class(class_name, module_name):
    # Get the class object itself from strings.
    loaded_module = importlib.import_module(module_name)
    try:
        class_ = getattr(loaded_module, class_name)
    except AttributeError as e:
        raise AttributeError("Module : {} has no class named : {}".format(
            module_name,
            class_name,
        ))
    return class_


# TODO: Rename runtime_config to runtime_environment and pass it through as a typed object, rather than unpacking it.
# TODO: Rename config to constructor_kwargs and config_defaults -> constructor_kwarg_default
# TODO: Improve error messages in this method. Since so much of our workflow is config-driven, this will be a *super* important part of DX.
def instantiate_class_from_config(config, runtime_config, config_defaults=None):
    """Build a GE class from configuration dictionaries."""

    if config_defaults is None:
        config_defaults = {}

    config = copy.deepcopy(config)

    module_name = config.pop("module_name", None)
    if module_name is None:
        try:
            module_name = config_defaults.pop("module_name")
        except KeyError as e:
            raise KeyError("Neither config : {} nor config_defaults : {} contains a module_name key.".format(
                config, config_defaults,
            ))
    else:
        # Pop the value without using it, to avoid sending an unwanted value to the config_class
        config_defaults.pop("module_name", None)

    class_name = config.pop("class_name", None)
    if class_name is None:
    # TODO : Trap this error and throw an informative message
        try:
            class_name = config_defaults.pop("class_name")
        except KeyError as e:
            raise KeyError("Neither config : {} nor config_defaults : {} contains a class_name key.".format(
                config, config_defaults,
            ))
    else:
        # Pop the value without using it, to avoid sending an unwanted value to the config_class
        config_defaults.pop("class_name", None)

    class_ = load_class(class_name=class_name, module_name=module_name)

    config_with_defaults = copy.deepcopy(config_defaults)
    config_with_defaults.update(config)
    config_with_defaults.update(runtime_config)

    try:
        class_instance = class_(**config_with_defaults)
    except TypeError as e:
        raise TypeError("Couldn't instantiate class : {} with config : \n\t{}\n \n".format(
            class_name,
            format_dict_for_error_message(config_with_defaults)
        ) + str(e))

    return class_instance


def format_dict_for_error_message(dict_):
    # TODO : Tidy this up a bit. Indentation isn't fully consistent.

    return '\n\t'.join('\t\t'.join((str(key), str(dict_[key]))) for key in dict_)


def substitute_config_variable(template_str, config_variables_dict):
    """
    This method takes a string, and if it is of the form ${SOME_VARIABLE} or $SOME_VARIABLE,
    returns the value of SOME_VARIABLE, otherwise returns the string unchanged.

    If the environment variable SOME_VARIABLE is set, the method uses its value for substitution.
    If it is not set, the value of SOME_VARIABLE is looked up in the config variables store (file).
    If it is not found there, the input string is returned as is.

    :param template_str: a string that might or might not be of the form ${SOME_VARIABLE}
            or $SOME_VARIABLE
    :param config_variables_dict: a dictionary of config variables. It is loaded from the
            config variables store (by default, "uncommitted/config_variables.yml file)
    :return:
    """
    if template_str is None:
        return template_str
    try:
        match = re.search(r'^\$\{(.*?)\}$', template_str) or re.search(r'^\$([_a-z][_a-z0-9]*)$', template_str)
    except TypeError:
        # If the value is not a string (e.g., a boolean), we should return it as is
        return template_str

    if match:
        config_variable_value = os.getenv(match.group(1))
        if not config_variable_value:
            config_variable_value = config_variables_dict.get(match.group(1))
            if not config_variable_value:
                config_variable_value = template_str
    else:
        config_variable_value = template_str

    return config_variable_value


def substitute_all_config_variables(data, replace_variables_dict):
    """
    Substitute all config variables of the form ${SOME_VARIABLE} in a dictionary-like
    config object for their values.

    The method traverses the dictionary recursively.

    :param data:
    :param replace_variables_dict:
    :return: a dictionary with all the variables replaced with their values
    """
    if isinstance(data, dict) or isinstance(data, OrderedDict):
        return {k : substitute_all_config_variables(v, replace_variables_dict) for k, v in data.items()}
    elif isinstance(data, list):
        return [substitute_all_config_variables(v, replace_variables_dict) for v in data]
    return substitute_config_variable(data, replace_variables_dict)


import pytest

from great_expectations.exceptions import MissingTopLevelConfigKeyError
from great_expectations.util import (
    gen_directory_tree_str,
)
from great_expectations.data_context.types import (
    ValidationResultIdentifier,
    SiteSectionIdentifier,
)
from great_expectations.data_context.store import (
    HtmlSiteStore
)

def test_HtmlSiteStore(tmp_path_factory):

    path = str(tmp_path_factory.mktemp('test_HtmlSiteStore_with_FixedLengthTupleFileSystemStoreBackend__dir'))
    project_path = str(tmp_path_factory.mktemp('my_dir'))

    my_store = HtmlSiteStore(
        root_directory = path,
        base_directory = "my_store",
    )

    with pytest.raises(TypeError):
        my_store.get("not_a_ValidationResultIdentifier")

    with pytest.raises(MissingTopLevelConfigKeyError):
        my_store.get(ValidationResultIdentifier(**{}))
    
    ns_1 = SiteSectionIdentifier(
        site_section_name="my_site_section",
        resource_identifier=ValidationResultIdentifier(
            from_string="ValidationResultIdentifier.a.b.c.quarantine.prod-100"
        )
    )
    my_store.set(ns_1, "aaa")
    # assert my_store.get(ns_1) == "aaa"

    ns_2 = SiteSectionIdentifier(
        site_section_name="my_site_section",
        resource_identifier=ValidationResultIdentifier(
            from_string="ValidationResultIdentifier.a.b.c.quarantine.prod-20"
        )
    )
    my_store.set(ns_2, "bbb")
    # assert my_store.get(ns_2) == {"B": "bbb"}

    print(my_store.list_keys())
    assert set(my_store.list_keys()) == set([
        ns_1,
        ns_2,
    ])

    print(gen_directory_tree_str(path))
    assert gen_directory_tree_str(path) == """\
test_HtmlSiteStore_with_FixedLengthTupleFileSystemStoreBackend__dir0/
    my_store/
        validations/
            prod-100/
                a/
                    b/
                        c/
                            quarantine.html
            prod-20/
                a/
                    b/
                        c/
                            quarantine.html
"""



import logging
logger = logging.getLogger(__name__)

import importlib
from six import string_types
import copy
import json
import os

import pandas as pd

from ..types.base_resource_identifiers import (
    DataContextKey,
)
from great_expectations.data_context.types.resource_identifiers import (
    ExpectationSuiteIdentifier,
    ValidationResultIdentifier,
    SiteSectionIdentifier,
)
from .store import (
    ReadWriteStore,
)
from great_expectations.data_context.util import (
    instantiate_class_from_config
)

class NamespacedReadWriteStore(ReadWriteStore):
    
    def __init__(self,
        store_backend,
        root_directory,
        serialization_type="json"
    ):
        self.store_backend = self._init_store_backend(
            copy.deepcopy(store_backend),
            runtime_config={
                "root_directory": root_directory
            }
        )
        self.root_directory = root_directory
        self.serialization_type = serialization_type

    def _init_store_backend(self, store_backend_config, runtime_config):
        if store_backend_config["class_name"] == "FixedLengthTupleFilesystemStoreBackend":
            config_defaults = {
                "key_length" : 5,
                "module_name" : "great_expectations.data_context.store",
            }
        elif store_backend_config["class_name"] == "FixedLengthTupleS3StoreBackend":
            config_defaults = {
                "key_length": 5, # NOTE: Eugene: 2019-09-06: ???
                "module_name": "great_expectations.data_context.store",
            }
        else:
            config_defaults = {
                "module_name" : "great_expectations.data_context.store",
            }

        return instantiate_class_from_config(
            config=store_backend_config,
            runtime_config=runtime_config,
            config_defaults=config_defaults,
        )

    def _get(self, key):
        self._validate_key(key)

        key_tuple = self._convert_resource_identifier_to_tuple(key)
        return self.store_backend.get(key_tuple)

    def _set(self, key, serialized_value):
        self._validate_key(key)

        key_tuple = self._convert_resource_identifier_to_tuple(key)
        return self.store_backend.set(key_tuple, serialized_value)

    def list_keys(self):
        return [self._convert_tuple_to_resource_identifier(key) for key in self.store_backend.list_keys()]

    def has_key(self, key):
        # NOTE: This is not efficient
        return key in self.list_keys()

    def _convert_resource_identifier_to_tuple(self, key):
        # TODO : Optionally prepend a source_id (the frontend Store name) to the tuple.

        list_ = []
        list_ += self._convert_resource_identifier_to_list(key)

        return tuple(list_)

    def _convert_resource_identifier_to_list(self, key):
        # The logic in this function is recursive, so it can't return a tuple

        list_ = []
        #Fetch keys in _key_order to guarantee tuple ordering in both python 2 and 3
        for key_name in key._key_order:
            key_element = key[key_name]
            if isinstance( key_element, DataContextKey ):
                list_ += self._convert_resource_identifier_to_list(key_element)
            else:
                list_.append(key_element)

        return list_

    def _convert_tuple_to_resource_identifier(self, tuple_):
        new_identifier = self.key_class(*tuple_)
        return new_identifier

    def _validate_key(self, key):
        if not isinstance(key, self.key_class):
            raise TypeError("key: {!r} must be a {}, not {!r}".format(
                key,
                self.key_class,
                type(key),
            ))


class ExpectationStore(NamespacedReadWriteStore):
    # Note : As of 2019/09/06, this method is untested.
    # It shares virtually all of its business logic with ValidationStore, which is under test.

    def _init_store_backend(self, store_backend_config, runtime_config):
        self.key_class = ExpectationSuiteIdentifier

        if store_backend_config["class_name"] == "FixedLengthTupleFilesystemStoreBackend":
            config_defaults = {
                "key_length" : 4,
                "module_name" : "great_expectations.data_context.store",
                "filepath_template": '{0}/{1}/{2}/{3}.json',
            }
        else:
            config_defaults = {
                "module_name" : "great_expectations.data_context.store",
            }

        return instantiate_class_from_config(
            config=store_backend_config,
            runtime_config=runtime_config,
            config_defaults=config_defaults,
        )

class ValidationResultStore(NamespacedReadWriteStore):
    
    def _init_store_backend(self, store_backend_config, runtime_config):
        self.key_class = ValidationResultIdentifier

        if store_backend_config["class_name"] == "FixedLengthTupleFilesystemStoreBackend":
            config_defaults = {
                "key_length" : 5,
                "module_name" : "great_expectations.data_context.store",
            }
        else:
            config_defaults = {
                "module_name" : "great_expectations.data_context.store",
            }

        return instantiate_class_from_config(
            config=store_backend_config,
            runtime_config=runtime_config,
            config_defaults=config_defaults,
        )


class HtmlSiteStore(NamespacedReadWriteStore):

    def __init__(self,
        root_directory,
        base_directory,
        serialization_type=None,
    ):
        # Each key type gets its own backend.
        # If backends were DB connections, this could be inefficient, but it doesn't much matter for filepaths.
        # One thing to watch for is reversibility of keys.
        # If several types are being writtten to overlapping directories, we could get collisions.
        self.store_backends = {
            ExpectationSuiteIdentifier : instantiate_class_from_config(
                config = {
                    "module_name" : "great_expectations.data_context.store",
                    "class_name" : "FixedLengthTupleFilesystemStoreBackend",
                    "key_length" : 4,
                    "base_directory" : base_directory,
                    "filepath_template" : 'expectations/{0}/{1}/{2}/{3}.html',
                },
                runtime_config={
                    "root_directory": root_directory
                }
            ),
            ValidationResultIdentifier : instantiate_class_from_config(
                config = {
                    "module_name" : "great_expectations.data_context.store",
                    "class_name" : "FixedLengthTupleFilesystemStoreBackend",
                    "key_length" : 5,
                    "base_directory" : base_directory,
                    "filepath_template" : 'validations/{4}/{0}/{1}/{2}/{3}.html',
                },
                runtime_config={
                    "root_directory": root_directory
                }
            ),
        }

        self.root_directory = root_directory
        self.serialization_type = serialization_type
        self.base_directory = base_directory

        # NOTE: Instead of using the filesystem as the source of record for keys,
        # this class trackes keys separately in an internal set.
        # This means that keys are stored for a specific session, but can't be fetched after the original
        # HtmlSiteStore instance leaves scope.
        # Doing it this way allows us to prevent namespace collisions among keys while still having multiple
        # backends that write to the same directory structure.
        # It's a pretty reasonable way for HtmlSiteStore to do its job---you just ahve to remember that it
        # can't necessarily set and list_keys like most other Stores.
        self.keys = set()

    def _get(self, key):
        self._validate_key(key)

        key_tuple = self._convert_resource_identifier_to_tuple(key.resource_identifier)
        return self.store_backends[
            type(key.resource_identifier)
        ].get(key_tuple)

    def _set(self, key, serialized_value):
        self._validate_key(key)

        self.keys.add(key)

        key_tuple = self._convert_resource_identifier_to_tuple(key.resource_identifier)
        return self.store_backends[
            type(key.resource_identifier)
        ].set(key_tuple, serialized_value)

    def _validate_key(self, key):
        if not isinstance(key, SiteSectionIdentifier):
            raise TypeError("key: {!r} must a SiteSectionIdentifier, not {!r}".format(
                key,
                type(key),
            ))

        for key_class in self.store_backends.keys():
            if isinstance(key.resource_identifier, key_class):
                return

        # The key's resource_identifier didn't match any known key_class
        raise TypeError("resource_identifier in key: {!r} must one of {}, not {!r}".format(
            key,
            set(self.store_backends.keys()),
            type(key),
        ))

    def list_keys(self):
        return list(self.keys)

    def write_index_page(self, page):
        # NOTE: This method is a temporary hack.
        # Status as of 2019/09/09: this method is backward compatible against the previous implementation of site_builder
        # However, it doesn't support backend pluggability---only implementation in a local filesystem.
        # Also, if/when we want to support index pages at multiple levels of nesting, we'll need to extend.
        # 
        # Properly speaking, what we need is a class of BackendStore that can accomodate this...
        # It's tricky with the current stores, sbecause the core get/set logic depends so strongly on fixed-length keys.
        index_page_path = os.path.join(self.root_directory, self.base_directory, "index.html")
        with open(index_page_path, "w") as file_:
            file_.write(page)
        return index_page_path



from .actions import (
    BasicValidationAction,
    NamespacedValidationAction,
    NoOpAction,
    StoreAction,
    ExtractAndStoreEvaluationParamsAction,
    SlackNotificationAction
)

from .validation_operators import (
    ValidationOperator,
    PerformActionListValidationOperator,
    RunWarningAndFailureExpectationSuitesValidationOperator
)

from .util import *




# -*- coding: utf-8 -*-
from great_expectations import rtd_url_ge_version

PROJECT_HELP_COMMENT = """
# Welcome to Great Expectations! Always know what to expect from your data.
# 
# This project configuration file allows you to define datasources, generators,
# integrations, and other things to help you crush pipeline debt.
#
# This file is intended to be committed to your source control repo.

# For more help configuring great expectations see the documentation at:
# https://docs.greatexpectations.io/en/""" + rtd_url_ge_version + """/core_concepts/data_context.html#configuration
# or join our slack channel: http://greatexpectations.io/slack

# NOTE: GE uses the names of configured `datasources` and `generators` to manage
# how `expectations` and other artifacts are stored in the `expectations/` and 
# `datasources/` folders. If you need to rename an existing `datasource` or 
# `generator`, be sure to also update the relevant directory names.

config_version: 1

# Datasources tell Great Expectations where your data lives and how to get it.
# You can use the CLI command `great_expectations add-datasource` to help you
# add a new datasource.
#
# Read more here at https://docs.greatexpectations.io/en/""" + rtd_url_ge_version + """/features/datasource.html
datasources: {}

"""

CONFIG_VARIABLES_INTRO = """
# This config file supports variable substitution which enables two use cases:
#   - Secrets are kept out of committed files.
#   - Configuration parameters can change based on the environment. For example:
#     dev vs staging vs prod.
#
# When GE encounters substitution syntax (like the examples below) in the config 
# file it will attempt to replace the value of “my_key” with the value from an 
# environment variable “my_value” or a corresponding key read from the file
# specified using `config_variables_file_path`. An environment variable will
# always take precedence.
#   - `my_key: ${my_value}
#   - `my_key: $my_value`
#
# If the substitution value comes from the config variables file, it can be a
# simple (non-nested) value or a nested value such as a dictionary. If it comes
# from an environment variable, it must be a simple value.
"""

PROJECT_OPTIONAL_CONFIG_COMMENT = CONFIG_VARIABLES_INTRO + """
config_variables_file_path: uncommitted/config_variables.yml

# The plugins_directory will be added to your python path for custom modules
# used to override and extend Great Expectations.

plugins_directory: plugins/

# Stores are configurable places to store things like expectations, evaluation
# results, data docs, and more.
# 
# Three stores are required: expectations, profiling, and evaluation parameters,
# and must be named here. Additional stores can be configured below and used for
# data_docs, validation_operators or other purposes.
#
# Point the following required store names to the appropriate `stores` entry.

expectations_store_name: expectations_store
profiling_store_name: local_validation_result_store
evaluation_parameter_store_name: evaluation_parameter_store

stores:
  expectations_store:
    class_name: ExpectationStore
    store_backend:
      class_name: FixedLengthTupleFilesystemStoreBackend
      base_directory: expectations/

  local_validation_result_store:
    class_name: ValidationResultStore
    store_backend:
      class_name: FixedLengthTupleFilesystemStoreBackend
      base_directory: uncommitted/validations/
      filepath_template: '{4}/{0}/{1}/{2}/{3}.json'

  # Uncomment the lines below to enable s3 as a result store. When enabled,
  # validation results will be saved in the store according to `run id`.
  # For S3, ensure that appropriate credentials or assume_role permissions are
  # set where validation happens.
  
  # s3_validation_result_store:
  #   class_name: ValidationResultStore
  #   store_backend:
  #     class_name: FixedLengthTupleS3StoreBackend
  #     bucket: <my_bucket>
  #     prefix: <my_prefix>
  #     file_extension: json
  #     filepath_template: '{4}/{0}/{1}/{2}/{3}.{file_extension}'


  evaluation_parameter_store:
    # Evaluation Parameters enable dynamic expectations. Read more here:
    # https://docs.greatexpectations.io/en/""" + rtd_url_ge_version + """/reference/evaluation_parameters.html
    module_name: great_expectations.data_context.store
    class_name: EvaluationParameterStore
  
  fixture_validation_results_store:
    class_name: ValidationResultStore
    store_backend:
      class_name: FixedLengthTupleFilesystemStoreBackend
      base_directory: fixtures/validations
      filepath_template: '{4}/{0}/{1}/{2}/{3}.json'
  
  local_site_html_store:
    module_name: great_expectations.data_context.store
    class_name: HtmlSiteStore
    base_directory: uncommitted/documentation/local_site/

  team_site_html_store:
    module_name: great_expectations.data_context.store
    class_name: HtmlSiteStore
    base_directory: uncommitted/documentation/team_site/


validation_operators:
  # Read about validation operators at: https://docs.greatexpectations.io/en/""" + rtd_url_ge_version + """/guides/validation_operators.html
  default:
    class_name: PerformActionListValidationOperator
    action_list:
      - name: store_validation_result
        action:
          class_name: StoreAction
          target_store_name: local_validation_result_store
      - name: store_evaluation_params
        action:
          class_name: ExtractAndStoreEvaluationParamsAction
          target_store_name: evaluation_parameter_store
      # Uncomment the notify_slack action below to send notifications during evaluation
      # - name: notify_slack
      #   action:
      #     class_name: SlackNotificationAction
      #     slack_webhook: ${validation_notification_slack_webhook}
      #     notify_on: all
      #     renderer:
      #       module_name: great_expectations.render.renderer.slack_renderer
      #       class_name: SlackRenderer
    

# TODO : Remove the extra layer of yml nesting in v0.8:
data_docs:
  sites:
    local_site: # site name
    # “local_site” renders documentation for all the datasources in the project from GE artifacts in the local repo. 
    # The site includes expectation suites and profiling and validation results from uncommitted directory. 
    # Local site provides the convenience of visualizing all the entities stored in JSON files as HTML.

      # specify a whitelist here if you would like to restrict the datasources to document
      datasource_whitelist: '*'

      module_name: great_expectations.render.renderer.site_builder
      class_name: SiteBuilder
      target_store_name: local_site_html_store
      
      site_index_builder:
        class_name: DefaultSiteIndexBuilder
      
      site_section_builders:
          
        expectations:
          class_name: DefaultSiteSectionBuilder
          source_store_name: expectations_store
          renderer:
            module_name: great_expectations.render.renderer
            class_name: ExpectationSuitePageRenderer

        validations:
          class_name: DefaultSiteSectionBuilder
          source_store_name: local_validation_result_store
          run_id_filter:
            ne: profiling
          renderer:
            module_name: great_expectations.render.renderer
            class_name: ValidationResultsPageRenderer

        profiling:
          class_name: DefaultSiteSectionBuilder
          source_store_name: local_validation_result_store
          run_id_filter:
            eq: profiling
          renderer:
            module_name: great_expectations.render.renderer
            class_name: ProfilingResultsPageRenderer

    team_site:
    # "team_site" is meant to support the "shared source of truth for a team" use case. 
    # By default only the expectations section is enabled.
    #  Users have to configure the profiling and the validations sections (and the corresponding validations_store and profiling_store attributes based on the team's decisions where these are stored (a local filesystem or S3). 
    # Reach out on Slack (https://greatexpectations.io/slack>) if you would like to discuss the best way to configure a team site.

      # specify a whitelist here if you would like to restrict the datasources to document
      datasource_whitelist: '*'
      
      module_name: great_expectations.render.renderer.site_builder
      class_name: SiteBuilder
      target_store_name: team_site_html_store
      
      site_index_builder:
        class_name: DefaultSiteIndexBuilder
      
      site_section_builders:
          
        expectations:
          class_name: DefaultSiteSectionBuilder
          source_store_name: expectations_store
          renderer:
            module_name: great_expectations.render.renderer
            class_name: ExpectationSuitePageRenderer

"""

PROJECT_TEMPLATE = PROJECT_HELP_COMMENT + PROJECT_OPTIONAL_CONFIG_COMMENT

CONFIG_VARIABLES_COMMENT = CONFIG_VARIABLES_INTRO

CONFIG_VARIABLES_FILE_TEMPLATE = CONFIG_VARIABLES_INTRO


import logging
logger = logging.getLogger(__name__)

import json
import importlib
from collections import OrderedDict

from great_expectations.render.renderer import (
    SiteIndexPageRenderer
)
from great_expectations.render.view import (
    DefaultJinjaIndexPageView,
)
from great_expectations.data_context.types import (
    ValidationResultIdentifier,
    ExpectationSuiteIdentifier,
    DataAssetIdentifier,
    NormalizedDataAssetName,
)
from great_expectations.data_context.util import (
    instantiate_class_from_config,
)

from great_expectations.data_context.store.namespaced_read_write_store import (
    SiteSectionIdentifier,
)

class SiteBuilder(object):
    """SiteBuilder builds data documentation for the project defined by a DataContext.

    A data documentation site consists of HTML pages for expectation suites, profiling and validation results, and
    an index.html page that links to all the pages.

    The exact behavior of SiteBuilder is controlled by configuration in the DataContext's great_expectations.yml file.

    Users can specify
    * which datasources to document (by default, all)
    * whether to include expectations, validations and profiling results sections
    * where the expectations and validations should be read from (filesystem or S3)
    * where the HTML files should be written (filesystem or S3)
    * which renderer and view class should be used to render each section

    Here is an example of config for a site:

    local_site:
        
        class_name: SiteBuilder
        target_store_name: local_site_html_store

        site_index_builder:
            class_name: DefaultSiteIndexBuilder

        # Verbose version:
        # index_builder:
        #     module_name: great_expectations.render.builder
        #     class_name: DefaultSiteIndexBuilder
        #     renderer:
        #         module_name: great_expectations.render.renderer
        #         class_name: SiteIndexPageRenderer
        #     view:
        #         module_name: great_expectations.render.view
        #         class_name: DefaultJinjaIndexPageView

        site_section_builders:
            # Minimal specification
            expectations:
                class_name: DefaultSiteSectionBuilder
                source_store_name: expectation_store
            renderer:
                module_name: great_expectations.render.renderer
                class_name: ExpectationSuitePageRenderer
                    
            # More verbose specification with optional arguments
            validations:
                module_name: great_expectations.data_context.render
                class_name: DefaultSiteSectionBuilder
                source_store_name: local_validation_store
                renderer:
                    module_name: great_expectations.render.renderer
                    class_name: SiteIndexPageRenderer
                view:
                    module_name: great_expectations.render.view
                    class_name: DefaultJinjaIndexPageView

        # specify a whitelist here if you would like to restrict the datasources to document
        datasource_whitelist:

    """

    def __init__(self,
        data_context,
        target_store_name,
        site_index_builder,
        site_section_builders,
        datasource_whitelist=None
    ):
        self.data_context = data_context
        self.target_store_name = target_store_name

        # the site config may specify the list of datasource names to document.
        # if the config property is absent or is *, treat as "all"
        self.datasource_whitelist = datasource_whitelist
        if not self.datasource_whitelist or self.datasource_whitelist == '*':
            self.datasource_whitelist = [datasource['name'] for datasource in data_context.list_datasources()]

        self.site_index_builder = instantiate_class_from_config(
            config=site_index_builder,
            runtime_config={
                "data_context": data_context,
                "target_store_name": target_store_name,
            },
            config_defaults={
                "name": "site_index_builder",
                "module_name": "great_expectations.render.renderer.site_builder"
            }
        )

        self.site_section_builders = {}
        for site_section_name, site_section_config in site_section_builders.items():
            self.site_section_builders[site_section_name] = instantiate_class_from_config(
                config=site_section_config,
                runtime_config={
                    "data_context": data_context,
                    "target_store_name": target_store_name,
                },
                config_defaults={
                    "name": site_section_name,
                    "module_name": "great_expectations.render.renderer.site_builder"
                }
            )

    def build(self):
        for site_section, site_section_builder in self.site_section_builders.items():
            site_section_builder.build(datasource_whitelist=self.datasource_whitelist)
        
        return self.site_index_builder.build()


class DefaultSiteSectionBuilder(object):

    def __init__(self,
        name,
        data_context,
        target_store_name,
        source_store_name,
        run_id_filter=None, #NOTE: Ideally, this would allow specification of ANY element (or combination of elements) within an ID key
        renderer=None,
        view=None,
    ):
        self.name = name
        self.source_store = data_context.stores[source_store_name]
        self.target_store = data_context.stores[target_store_name]
        self.run_id_filter = run_id_filter

        # TODO : Push conventions for configurability down to renderers and views.
        # Until then, they won't be configurable, and defaults will be hard.
        if renderer == None:
            renderer = {
                "module_name": "great_expectations.render.renderer",
                # "class_name": "SiteIndexPageRenderer",
            }
        renderer_module = importlib.import_module(renderer.pop("module_name"))
        self.renderer_class = getattr(renderer_module, renderer.pop("class_name"))

        if view == None:
            view = {
                "module_name": "great_expectations.render.view",
                "class_name": "DefaultJinjaPageView",
            }
        view_module = importlib.import_module(view.pop("module_name"))
        self.view_class = getattr(view_module, view.pop("class_name"))

    def build(self, datasource_whitelist):
        for resource_key in self.source_store.list_keys():

            if self.run_id_filter:
                if not self._resource_key_passes_run_id_filter(resource_key):
                    continue
                    
            if not self._resource_key_passes_datasource_whitelist(resource_key, datasource_whitelist):
                continue
                
            resource = self.source_store.get(resource_key)

            if type(resource_key) is ExpectationSuiteIdentifier:
                expectation_suite_name = resource_key.expectation_suite_name
                data_asset_name = resource_key.data_asset_name.generator_asset
                logger.info(
                    "        Rendering expectation suite {} for data asset {}".format(
                        expectation_suite_name,
                        data_asset_name
                    ))
            elif type(resource_key) is ValidationResultIdentifier:
                data_asset_name = resource_key.expectation_suite_identifier.data_asset_name.generator_asset
                run_id = resource_key.run_id
                expectation_suite_name = resource_key.expectation_suite_identifier.expectation_suite_name
                if run_id == "profiling":
                    logger.info("        Rendering profiling for data asset {}".format(data_asset_name))
                else:
                    
                    logger.info("        Rendering validation: run id: {}, suite {} for data asset {}".format(run_id,
                                                                                                              expectation_suite_name,
                                                                                                              data_asset_name))

            # TODO : Typing resources is SUPER important for usability now that we're slapping configurable renders together with arbitrary stores.
            rendered_content = self.renderer_class.render(resource)
            viewable_content = self.view_class.render(rendered_content)

            self.target_store.set(
                SiteSectionIdentifier(
                    site_section_name=self.name,
                    resource_identifier=resource_key,
                ),
                viewable_content
            )

    def _resource_key_passes_datasource_whitelist(self, resource_key, datasource_whitelist):
        if type(resource_key) is ExpectationSuiteIdentifier:
            datasource = resource_key.data_asset_name.datasource
        elif type(resource_key) is ValidationResultIdentifier:
            datasource = resource_key.expectation_suite_identifier.data_asset_name.datasource
        return datasource in datasource_whitelist
    
    def _resource_key_passes_run_id_filter(self, resource_key):
        if type(resource_key) == ValidationResultIdentifier:
            run_id = resource_key.run_id
        else:
            raise TypeError("run_id_filter filtering is only implemented for ValidationResultResources.")

        if self.run_id_filter.get("eq"):
            return self.run_id_filter.get("eq") == run_id

        elif self.run_id_filter.get("ne"):
            return self.run_id_filter.get("ne") != run_id


class DefaultSiteIndexBuilder(object):

    def __init__(self,
        name,
        data_context,
        target_store_name,
        renderer=None,
        view=None,
    ):
        # NOTE: This method is almost idenitcal to DefaultSiteSectionBuilder
        self.name = name
        self.data_context = data_context
        self.target_store = data_context.stores[target_store_name]

        # TODO : Push conventions for configurability down to renderers and views.
        # Until then, they won't be configurable, and defaults will be hard.
        if renderer == None:
            renderer = {
                "module_name": "great_expectations.render.renderer",
                "class_name": "SiteIndexPageRenderer",
            }
        renderer_module = importlib.import_module(renderer.pop("module_name"))
        self.renderer_class = getattr(renderer_module, renderer.pop("class_name"))

        if view == None:
            view = {
                "module_name" : "great_expectations.render.view",
                "class_name" : "DefaultJinjaIndexPageView",
            }
        view_module = importlib.import_module(view.pop("module_name"))
        self.view_class = getattr(view_module, view.pop("class_name"))

    def add_resource_info_to_index_links_dict(self,
                                              data_context,
                                              index_links_dict,
                                              data_asset_name,
                                              datasource,
                                              generator,
                                              generator_asset,
                                              expectation_suite_name,
                                              section_name,
                                              run_id=None,
                                              validation_success=None
                                              ):
        if not datasource in index_links_dict:
            index_links_dict[datasource] = OrderedDict()
    
        if not generator in index_links_dict[datasource]:
            index_links_dict[datasource][generator] = OrderedDict()
    
        if not generator_asset in index_links_dict[datasource][generator]:
            index_links_dict[datasource][generator][generator_asset] = {
                'profiling_links': [],
                'validations_links': [],
                'expectations_links': []
            }
    
        if run_id:
            base_path = "validations/" + run_id
        else:
            base_path = "expectations"
    
        index_links_dict[datasource][generator][generator_asset][section_name + "_links"].append(
            {
                "full_data_asset_name": data_asset_name,
                "expectation_suite_name": expectation_suite_name,
                "filepath": data_context._get_normalized_data_asset_name_filepath(
                    data_asset_name,
                    expectation_suite_name,
                    base_path=base_path,
                    file_extension=".html"
                ),
                "source": datasource,
                "generator": generator,
                "asset": generator_asset,
                "run_id": run_id,
                "validation_success": validation_success
            }
        )
    
        return index_links_dict

    def build(self):
        # Loop over sections in the HtmlStore
        logger.debug("DefaultSiteIndexBuilder.build")

        target_store = self.target_store
        resource_keys = target_store.list_keys()
        index_links_dict = OrderedDict()

        for key in resource_keys:
            key_resource_identifier = key.resource_identifier
            
            if type(key_resource_identifier) == ExpectationSuiteIdentifier:
                self.add_resource_info_to_index_links_dict(
                    data_context=self.data_context,
                    index_links_dict=index_links_dict,
                    data_asset_name=key_resource_identifier.data_asset_name.to_string(
                        include_class_prefix=False,
                        separator=self.data_context.data_asset_name_delimiter
                    ),
                    datasource=key_resource_identifier.data_asset_name.datasource,
                    generator=key_resource_identifier.data_asset_name.generator,
                    generator_asset=key_resource_identifier.data_asset_name.generator_asset,
                    expectation_suite_name=key_resource_identifier.expectation_suite_name,
                    section_name=key.site_section_name
                )
            elif type(key_resource_identifier) == ValidationResultIdentifier:
                data_asset_name = key_resource_identifier.expectation_suite_identifier.data_asset_name.to_string(
                        include_class_prefix=False,
                        separator=self.data_context.data_asset_name_delimiter
                    )
                expectation_suite_name = key_resource_identifier.expectation_suite_identifier.expectation_suite_name
                run_id = key_resource_identifier.run_id
                validation = self.data_context.get_validation_result(
                    data_asset_name=data_asset_name,
                    expectation_suite_name=expectation_suite_name,
                    validations_store_name="local_validation_result_store",
                    run_id=run_id
                )
                
                validation_success = validation.get("success")
                
                self.add_resource_info_to_index_links_dict(
                    data_context=self.data_context,
                    index_links_dict=index_links_dict,
                    data_asset_name=key_resource_identifier.expectation_suite_identifier.data_asset_name.to_string(
                        include_class_prefix=False,
                        separator=self.data_context.data_asset_name_delimiter
                    ),
                    datasource=key_resource_identifier.expectation_suite_identifier.data_asset_name.datasource,
                    generator=key_resource_identifier.expectation_suite_identifier.data_asset_name.generator,
                    generator_asset=key_resource_identifier.expectation_suite_identifier.data_asset_name.generator_asset,
                    expectation_suite_name=expectation_suite_name,
                    section_name=key.site_section_name,
                    run_id=run_id,
                    validation_success=validation_success
                )

        rendered_content = self.renderer_class.render(index_links_dict)
        viewable_content = self.view_class.render(rendered_content)

        return (
            self.target_store.write_index_page(viewable_content),
            index_links_dict
        )